/**
 * @file	gpio.c
 * @author	Robert Steigemann
 * @date	15 January 2017
 * @brief	Library for GPIO, buttons, and rotary encoder
 *
 * @license creative commons license CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>

#include "gpio.h"

#define SHORT_DURATION    5
#define LONG_DURATION	100
#define MAX_DURATION    127

static volatile int8_t encoder = 0; 

// Byte for each button is splitted in two parts:
// MSB marks if button is pressed (0) or released (1).
// Bit 0 to 6 counts how long the button is/was pressed.
static uint8_t buttonCtr[BUTTON_MAX] = {0};

static bool stb_sw_1 = false;
static bool stb_sw_2 = false;


bool gpio_get(gpio_item_t item)
{
	switch (item)
	{
	case ITEM_BUTTON_ROT_ENC:
		return ( (PORTE.IN & (1 << 1)) != 0 );
		
	case ITEM_BUTTON_1:
		return ( (PORTR.IN & (1 << 0)) != 0 );
	
	case ITEM_BUTTON_2:
		return ( (PORTR.IN & (1 << 1)) != 0 );
	
	case ITEM_BUTTON_3:
		return ( (PORTE.IN & (1 << 2)) != 0 );
	
	case ITEM_STB_SW_1:
		return ( (PORTC.IN & (1 << 2)) != 0 );
	
	case ITEM_STB_SW_2:
		return ( (PORTC.IN & (1 << 1)) != 0 );
	
	case ITEM_ENC_A:
		return ( (PORTE.IN & (1 << 3)) != 0 );
	
	case ITEM_ENC_B:
		return ( (PORTE.IN & (1 << 0)) != 0 );
		
	default:
		return false;
	}	
}

void gpio_poll(void)
{
	if ( !gpio_get(ITEM_BUTTON_1) )			// switch 1 pressed
	{
		if ( buttonCtr[BUTTON_1] >= 0x80 )	// MSB is set = first execution after button is pressed
		{
			buttonCtr[BUTTON_1] = 0;		// Start counting from the bottom
		}
		
		if ( buttonCtr[BUTTON_1] < 127 )	// Button count is lower maximum
		{
			buttonCtr[BUTTON_1] ++;			// Count up
		}
	}
	else									// switch is not pressed
	{
		buttonCtr[BUTTON_1] |= 0x80;		// set MSB, count value remains in bit 0 to 6	
	}
	
	if ( !gpio_get(ITEM_BUTTON_2) )			// switch 2 pressed
	{
		if ( buttonCtr[BUTTON_2] >= 0x80 )	
		{
			buttonCtr[BUTTON_2] = 0;		
		}
		
		if ( buttonCtr[BUTTON_2] < 127 )	
		{
			buttonCtr[BUTTON_2] ++;			
		}
	}
	else									
	{
		buttonCtr[BUTTON_2] |= 0x80;		
	}
	
	if ( !gpio_get(ITEM_BUTTON_3) )			// switch 3 pressed
	{
		if ( buttonCtr[BUTTON_3] >= 0x80 )	
		{
			buttonCtr[BUTTON_3] = 0;		
		}
		
		if ( buttonCtr[BUTTON_3] < 127 )	
		{
			buttonCtr[BUTTON_3] ++;			
		}
	}
	else									
	{
		buttonCtr[BUTTON_3] |= 0x80;		
	}
	
	if ( !gpio_get(ITEM_BUTTON_ROT_ENC) )	// switch enc pressed
	{
		if ( buttonCtr[BUTTON_ROT_ENC] >= 0x80 )
		{
			buttonCtr[BUTTON_ROT_ENC] = 0;
		}
		
		if ( buttonCtr[BUTTON_ROT_ENC] < 127 )
		{
			buttonCtr[BUTTON_ROT_ENC] ++;
		}
	}
	else
	{
		buttonCtr[BUTTON_ROT_ENC] |= 0x80;
	}
}


bool gpio_checkButton(gpio_button_t button, gpio_presstime_t duration)
{
	if (duration == SHORT)					// check if short press happened
	{
		// button was pressed in the past and not to long and not to short
		if (    ((buttonCtr[button] & 0x80) != 0)
		     && ((buttonCtr[button] & 0x7F) < LONG_DURATION )
			 && ((buttonCtr[button] & 0x7F) > SHORT_DURATION ) )
		{
			buttonCtr[button] = 0;			// reset button count
			return true;
		} 
		else
		{
			return false;
		}		
	} 
	else if (duration == LONG)
	{
		// button is pressed and longer then long duration
		if (    ((buttonCtr[button] & 0x80) == 0)
		     && ((buttonCtr[button] & 0x7F) >= LONG_DURATION )
			 && ((buttonCtr[button] & 0x7F) < MAX_DURATION ) )
		{
			buttonCtr[button] = MAX_DURATION;			// reset button count
			return true;
		} 
		else
		{
			return false;
		}
	}
	else
	{
		return false;
	}
}

bool gpio_iron_taken(void)
{
	bool taken = false;
	
	if ( gpio_get(ITEM_STB_SW_1) && !stb_sw_1 )
	{
		stb_sw_1 = true;
		taken = true;
	}
	
	if ( !gpio_get(ITEM_STB_SW_2) && stb_sw_2 )
	{
		stb_sw_1 = false;
		taken = true;
	}
	
	return taken;	
}

bool gpio_iron_inserted(void)
{
	bool taken = false;
	
	if ( !gpio_get(ITEM_STB_SW_1) && stb_sw_1 )
	{
		stb_sw_1 = false;
		taken = true;
	}
	
	if ( gpio_get(ITEM_STB_SW_2) && !stb_sw_2 )
	{
		stb_sw_1 = true;
		taken = true;
	}
	
	return taken;	
}

void gpio_init(void)
{
	/*
		PA0 - GNDA
		PA1 - GNDA
		PA2 - GNDA
		PA3 - GNDA
		PA4 - 245 mV, negative input
		PA5 - GNDA
		PA6 - sensor, positive input
		PA7 - GNDA	
	*/		
	PORTA.DIR = 0x00;
	// Disable input buffer, because pins are used for A/D only
	PORTA.PIN0CTRL = PORT_ISC_INPUT_DISABLE_gc;	
	PORTA.PIN1CTRL = PORT_ISC_INPUT_DISABLE_gc;
	PORTA.PIN2CTRL = PORT_ISC_INPUT_DISABLE_gc;
	PORTA.PIN3CTRL = PORT_ISC_INPUT_DISABLE_gc;
	PORTA.PIN4CTRL = PORT_ISC_INPUT_DISABLE_gc;
	PORTA.PIN5CTRL = PORT_ISC_INPUT_DISABLE_gc;
	PORTA.PIN6CTRL = PORT_ISC_INPUT_DISABLE_gc;
	PORTA.PIN7CTRL = PORT_ISC_INPUT_DISABLE_gc;
	
	/*
		PB0 - n.c.
		PB1 - n.c.
		PB2 - n.c.
		PB3 - n.c.
	*/
	PORTB.DIR = 0x00;
	// Pins are not connected, so use pulldown to terminate pins
	PORTB.PIN0CTRL = PORT_OPC_PULLDOWN_gc;
	PORTB.PIN1CTRL = PORT_OPC_PULLDOWN_gc;
	PORTB.PIN2CTRL = PORT_OPC_PULLDOWN_gc;
	PORTB.PIN3CTRL = PORT_OPC_PULLDOWN_gc;
	
	/*
		PC0 - Solderiron PWM Output, OC0A
		PC1 - Standby Switch 2 Pin 3
		PC2 - Standby Switch 1 Pin 2 
		PC3 - n.c.
		PC4 - n.c.
		PC5 - n.c.
		PC6 - UART RXD, RXDC1
		PC7 - UART TXD, TXDC1
	*/
	PORTC.DIR = 0b10000001; 
	PORTC.OUT = 0x00;
	// switch inputs require pullup
	PORTC.PIN1CTRL = PORT_OPC_PULLUP_gc;
	PORTC.PIN2CTRL = PORT_OPC_PULLUP_gc;
	// Pins are not connected, so use pulldown to terminate pins
	PORTC.PIN3CTRL = PORT_OPC_PULLDOWN_gc;
	PORTC.PIN4CTRL = PORT_OPC_PULLDOWN_gc;
	PORTC.PIN5CTRL = PORT_OPC_PULLDOWN_gc;
	
	/*
		PD0 - GND
		PD1 - GND
		PD2 - /CS, OLED Display
		PD3 - /RES, OLED Display
		PD4 - D/C, OLED Display
		PD5 - MOSI, SPID, OLED Display
		PD6 - MISO, SPID, n.c.
		PD7 - SCK, SPID, OLED Display
	*/
	PORTD.DIR = 0b10111100;
	// make /CS high until it is configured by ssd1309 library
	PORTD.OUT = 0b00000100;
	// give MISO a pull down, because it is not connected		
	PORTD.PIN6CTRL = PORT_OPC_PULLDOWN_gc;
	
	/*
		PE0 - IncB
		PE1 - Switch/Inc
		PE2 - Switch/Inc
		PE3 - IncA
	*/
	PORTE.DIR = 0x00;	
	// switch inputs require pullup
	PORTE.PIN0CTRL = PORT_OPC_PULLUP_gc;
	PORTE.PIN1CTRL = PORT_OPC_PULLUP_gc;
	PORTE.PIN2CTRL = PORT_OPC_PULLUP_gc;
	PORTE.PIN3CTRL = PORT_OPC_PULLUP_gc | PORT_ISC_BOTHEDGES_gc;
	// Interrupt for INCA
	PORTE.INT0MASK = (1 << 3);				// pin 3 triggers interrupt
	PORTE.INTCTRL = PORT_INT0LVL_HI_gc;		// activate low level interrupt0
	// PORTE.INTFLAGS = PORT_INT0IF_bm;		// clear INT0 interrupt flag
	// Thats useless here because the capacitors on the encoder a and b lines start charging
	// after the pullups are activated above. If the encoder lines are on high impedance state
	// during boot (50% chance), they will become high in a few ms and will set the flag again.
	// As workaround I clear the flag at the end of init in the main loop.	
		
	/*
		PR0 - Switch
		PR1 - Switch
	*/
	PORTR.DIR = 0x00;	
	// switch inputs require pullup
	PORTR.PIN0CTRL = PORT_OPC_PULLUP_gc;
	PORTR.PIN1CTRL = PORT_OPC_PULLUP_gc;
	
	// call these functions one time to set initial state
	gpio_iron_taken();
	gpio_iron_inserted();
}


// Encoder service routine
ISR(PORTE_INT0_vect)
{
	switch(PORTE.IN & ((1<<0)|(1<<3)))
	{
		case 0:
		case ((1<<0)|(1<<3)):
		if ( encoder < INT8_MAX )
		{
			encoder ++;
		}
		break;
		
		default:
		if ( encoder > INT8_MIN )
		{
			encoder --;
		}
		break;
	}
}


// returns the covered steps (+/-) since last check
int8_t gpio_getRotEnc(void)
{
	int8_t tmp;
	
	cli();
		tmp = encoder;
		encoder = 0;
	sei();
	
	return tmp;
}
