/**
 * @file	main.c
 * @author	Robert Steigemann
 * @date	15 January 2017
 * @brief	Firmware for second edition soldering station
 *
 * @license creative commons license CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/
 */

//////////////////////////////////////////////////////////////////////////
/////////////////////            Includes           //////////////////////
//////////////////////////////////////////////////////////////////////////

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stddef.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

#include "Clockconfig.h"
#include "gpio.h"
#include "Serialport.h"
#include "ssd1309.h"
#include "tempsens.h"
#include "pwm.h"


//////////////////////////////////////////////////////////////////////////
/////////////////////              Defines          //////////////////////
//////////////////////////////////////////////////////////////////////////

#define VERSION "1.3"

#define EEPROM_START_CHECKER 0xF0F0F0F0
#define EEPROM_END_CHECKER   0x0A0A0A0C


//////////////////////////////////////////////////////////////////////////
/////////////////////           EEPROM Image        //////////////////////
//////////////////////////////////////////////////////////////////////////

uint32_t EEMEM eeprom_start_checker = EEPROM_START_CHECKER;

// PID controller values
float EEMEM eeprom_p_value = 1.0;
float EEMEM eeprom_i_value = 0.05;
float EEMEM eeprom_d_value = 6.0;

// station values
int16_t EEMEM eeprom_target_temp = 300;
int16_t EEMEM eeprom_quickselect_temp[3] = {300, 350, 450};
int16_t EEMEM eeprom_standby_temp = 150;
uint8_t EEMEM eeprom_standby_time = 10;		// in minutes	
solderType_t EEMEM eeprom_tool = WELLER_WP80;
uint8_t EEMEM eeprom_contrast = 255;

uint32_t EEMEM eeprom_end_checker = EEPROM_END_CHECKER;


//////////////////////////////////////////////////////////////////////////
////////////////////        global variables        //////////////////////
//////////////////////////////////////////////////////////////////////////

// custom graphics for OLED
const uint8_t PROGMEM degreeCelsius[] = { 0x0E, 0x1B, 0x11, 0x1B, 0x8E,0xE0, 0x70, 0x30, 0x30, 0x30, 0x30, 0xE0, 0xC0,
										   0x00, 0x00, 0x00, 0x00, 0x07, 0x1F, 0x38, 0x30, 0x30, 0x30, 0x30, 0x1C, 0x0C };

const uint8_t PROGMEM marker[] = { 0x00, 0x7F, 0x3E, 0x1C, 0x08, 0x00 };
	
const uint8_t PROGMEM arrowExit[] = { 0x1C, 0x10, 0x10, 0x54, 0x28, 0x10 };
	
const uint8_t PROGMEM arrowDown[] = { 0x08, 0x10, 0x20, 0x10, 0x08, 0x00 };
	
const uint8_t PROGMEM arrowUp[] = { 0x10, 0x08, 0x04, 0x08, 0x10, 0x00 };


// PID controller parameters
float p_value = 1.0;
float i_value = 0.05;
float d_value = 6.0;

// Standby parameters
int16_t standby_temp = 150;
uint8_t standby_time = 10;		// in minutes

// Solder parameters
solderType_t tool = WELLER_WP80;

// Display brightness
uint8_t contrast = 255;

//////////////////////////////////////////////////////////////////////////
////////////////      local function prototypes        ///////////////////
//////////////////////////////////////////////////////////////////////////

void gpio_init(void);
uint8_t calc_pid(uint16_t targetTemperature, float actualTemperature, bool logEnable);
void settings_menu(void);

//////////////////////////////////////////////////////////////////////////
/////////////////////////      main function     /////////////////////////
//////////////////////////////////////////////////////////////////////////

int main(void)
{
    // serial receiver
	char delimiter[] = " ,;";
	char rcvString[64], *ptr;
	
	// soldering station dataset
	tempsensState_t solder_status = TOOL_DISCONNECTED;
	float actual_temp = 0.0;
	int16_t target_temp = 300;
	int16_t quickselect_temp[3] = {300, 350, 450};
	uint8_t power = 0;
	bool standby = false;			// remember standby state
	bool serialLog = false;			// serial log output
			
	clk_init();
	gpio_init();
	tempsens_init();
	pwm_init();	
	serialport_init(115200);
	
	
	// read the eeprom values, if eeprom has correct values
	if ((eeprom_read_dword(&eeprom_start_checker) == EEPROM_START_CHECKER) && (eeprom_read_dword(&eeprom_end_checker) == EEPROM_END_CHECKER))
	{
		eeprom_read_block(&tool, &eeprom_tool, sizeof(tool));
		eeprom_read_block(&target_temp, &eeprom_target_temp, sizeof(target_temp));
		eeprom_read_block(quickselect_temp, eeprom_quickselect_temp, sizeof(quickselect_temp));
		eeprom_read_block(&standby_temp, &eeprom_standby_temp, sizeof(standby_temp));
		standby_time = eeprom_read_byte(&eeprom_standby_time);		
		p_value = eeprom_read_float(&eeprom_p_value);
		i_value = eeprom_read_float(&eeprom_i_value);
		d_value = eeprom_read_float(&eeprom_d_value);	
		contrast = eeprom_read_byte(&eeprom_contrast);	
	} 
	else // initialize eeprom
	{
		eeprom_write_dword(&eeprom_start_checker, EEPROM_START_CHECKER);
		
		eeprom_write_block(&tool, &eeprom_tool, sizeof(tool));
		eeprom_write_block(&target_temp, &eeprom_target_temp, sizeof(target_temp));
		eeprom_write_block(quickselect_temp, eeprom_quickselect_temp, sizeof(quickselect_temp));
		eeprom_write_block(&standby_temp, &eeprom_standby_temp, sizeof(standby_temp));
		eeprom_write_byte(&eeprom_standby_time, standby_time);
		eeprom_write_float(&eeprom_p_value, p_value);		
		eeprom_write_float(&eeprom_i_value, i_value);
		eeprom_write_float(&eeprom_d_value, d_value);
		eeprom_write_byte(&eeprom_contrast, contrast);
		
		eeprom_write_dword(&eeprom_end_checker, EEPROM_END_CHECKER);
	}
	
	ssd1309_init(true, contrast);
		
		
	// clear INT0 interrupt flag to avoid ghost increasing of target temp by 1		
	PORTE.INTFLAGS = PORT_INT0IF_bm;		
	PMIC_CTRL |= PMIC_HILVLEN_bm|PMIC_MEDLVLEN_bm|PMIC_LOLVLEN_bm;
	sei();
	
	serialport_printf("\r\rLS2 V: ");
	serialport_printf(VERSION);
	serialport_printf(" by RS-Elec\rrselec.de\r");	
	
	while (1)			// main loop
	{
		if (standby)
		{
			if (0 != gpio_getRotEnc())
			{
				standby = false;
			}
		} 
		else
		{
			// update target temp.
			target_temp += gpio_getRotEnc();
			if (target_temp < 50)
			{
				target_temp = 50;
			}
			if (target_temp > 450)
			{
				target_temp = 450;
			}
		}
		
		// iron has been taken
		if (gpio_iron_taken())
		{
			standby = false;
		}
				
		
		if (gpio_checkButton(BUTTON_1, SHORT))
		{
			target_temp = quickselect_temp[0];
			standby = false;
		}
		
		if (gpio_checkButton(BUTTON_2, SHORT))
		{
			target_temp = quickselect_temp[1];
			standby = false;
		}
		
		if (gpio_checkButton(BUTTON_3, SHORT))
		{
			target_temp = quickselect_temp[2];
			standby = false;
		}
		
		if (gpio_checkButton(BUTTON_ROT_ENC, SHORT))
		{
			if (standby)
			{
				standby = false;
			} 
			else
			{
				standby = true;
			}			
		}
		
		
		if (gpio_checkButton(BUTTON_1, LONG))
		{
			quickselect_temp[0] = target_temp;
			eeprom_update_block(&quickselect_temp[0], &eeprom_quickselect_temp[0], sizeof(quickselect_temp[0]));
		}
		
		if (gpio_checkButton(BUTTON_2, LONG))
		{
			quickselect_temp[1] = target_temp;
			eeprom_update_block(&quickselect_temp[1], &eeprom_quickselect_temp[1], sizeof(quickselect_temp[1]));
		}
		
		if (gpio_checkButton(BUTTON_3, LONG))
		{
			quickselect_temp[2] = target_temp;
			eeprom_update_block(&quickselect_temp[2], &eeprom_quickselect_temp[2], sizeof(quickselect_temp[2]));
		}
		
		if (gpio_checkButton(BUTTON_ROT_ENC, LONG))
		{
			// goto menu
			settings_menu();
		}
		
				
		// every 10 ms
		if (timer_10ms_tick())
		{
			solder_status = tempsens_measure();		// take a measure sample
			gpio_poll();							// poll switches
		}
		
		// update Display
		if (timer_100ms_tick())
		{
			if (solder_status == AVCC_OUT_OF_RANGE)
			{
				ssd1309_clear();
				ssd1309_print_P(PSTR("Fatal system error!"), 3, 0);				
			} 
			else
			{
				// draw the bar on the left side
				ssd1309_bargraph(power, 0, 7, 0, 8);
				
				if (standby)
				{
					// draw standby value
					ssd1309_print_P(PSTR("STANDBY:"), 0, 24);
					ssd1309_printf_bigDigit_P(1, 24, PSTR("%3i"), standby_temp);
				}
				else
				{
					// draw target value
					ssd1309_print_P(PSTR("TARGET: "), 0, 24);
					ssd1309_printf_bigDigit_P(1, 24, PSTR("%3i"), target_temp);
				}				
				ssd1309_picture(degreeCelsius, 1, 2, 69, 13);
								
				// draw shortcuts
				ssd1309_printf_P(1, 96, PSTR("%3iC"), quickselect_temp[0]);
				ssd1309_printf_P(4, 96, PSTR("%3iC"), quickselect_temp[1]);
				ssd1309_printf_P(7, 96, PSTR("%3iC"), quickselect_temp[2]);
				
				// draw actual value
				ssd1309_print_P(PSTR("ACTUAL:"), 4, 24);				
			 	ssd1309_picture(degreeCelsius, 5, 6, 69, 13);				
				if (solder_status == OK)
				{
					ssd1309_printf_bigDigit_P(5, 24, PSTR("%3.0f"), actual_temp);
				} 
				else if (solder_status == TOOL_DISCONNECTED)
				{
					ssd1309_printf_bigDigit_P(5, 24, PSTR("---"));
				}
				
				ssd1309_print_P(PSTR("TOOL:"), 7, 24);
				
				// draw tool name
				switch (tool)
				{
				case WELLER_WP80:
					ssd1309_print_P(PSTR("WP80 "), 7, 54);
					break;
				case WELLER_WSP80:
					ssd1309_print_P(PSTR("WSP80"), 7, 54);
					break;
				default:
					break;
				}
								
			}
		}
		
		
		if (timer_1s_tick())
		{
			// store target temp in eeprom
			static int16_t old_target_temp;
			static uint8_t storage_timer = 0;
			if (old_target_temp == target_temp)
			{
				if (storage_timer < UINT8_MAX)
				{
					storage_timer ++;
				}				
				if (storage_timer == 60)	// target temp for one minute unchanged
				{
					eeprom_update_block(&target_temp, &eeprom_target_temp, sizeof(target_temp));
				}
			} 
			else
			{
				storage_timer = 0;
				old_target_temp = target_temp;
			}
			
			// calculate PID
			if (solder_status == OK)
			{
				actual_temp = tempsens_get(tool);
				if (standby)
				{
					power = calc_pid((uint16_t)standby_temp, actual_temp, serialLog);	
				} 
				else
				{
					power = calc_pid((uint16_t)target_temp, actual_temp, serialLog);	
				}						
			} 
			else
			{
				power = 0;				
			}			
			pwm_set_power(power);			
		}		
		
		// check serial communication	   
		if (0 != serialport_getLastString((uint8_t *)rcvString, 64))
		{
			serialport_printf("%s -> ", rcvString);
			ptr = strtok(rcvString, delimiter);			
			
			if (0 == strcasecmp_P(ptr, PSTR("reset")))						// system reset
			{
				// destruct start checker to load default values at next reboot
				eeprom_write_dword(&eeprom_start_checker, 0);
				serialport_printf("OK\r");
				cli();						// disable interrupts
				CPU_CCP = 0xD8;				// disable write protection
				RST_CTRL = 1;				// do a software reset
				_delay_ms(1);				// wait until reset occurs
			}
			else if (0 == strcasecmp_P(ptr, PSTR("standby")))				// enable/disable standby
			{
				ptr = strtok(NULL, delimiter);
				if (ptr != NULL)
				{
					if (0 == strcasecmp_P(ptr, PSTR("true")))
					{
						standby = true;
						serialport_printf("OK\r");
					}
					else if (0 == strcasecmp_P(ptr, PSTR("false")))
					{
						standby = false;
						serialport_printf("OK\r");
					}
					else
					{
						serialport_printf("ERROR, true or false required\r");
					}					
				}
				else
				{
					serialport_printf("ERROR, true or false required\r");
				}
			} 
			else if (0 == strcasecmp_P(ptr, PSTR("target")))				// set target temp
			{
				ptr = strtok(NULL, delimiter);
				if (ptr != NULL)
				{
					int aValue = atoi(ptr);
					if ( (aValue < 50) || (aValue > 450) )
					{
						serialport_printf("ERROR, valid temperature value required (50...450)\r");
					} 
					else
					{
						target_temp = aValue;
						serialport_printf("OK\r");
					}					
				}
				else
				{
					serialport_printf("ERROR, valid temperature value required (50...450)\r");
				}
			}
			else if (0 == strcasecmp_P(ptr, PSTR("version")))				// print version
			{
				serialport_printf("OK\r");
				serialport_printf(VERSION);
				serialport_printf("\r");
			}
			else if (0 == strcasecmp_P(ptr, PSTR("actual")))				// get actual iron temperature
			{
				serialport_printf("OK\r");
				serialport_printf("%5.1f\r", actual_temp);
			}
			else if (0 == strcasecmp_P(ptr, PSTR("log")))				// enable logging
			{
				ptr = strtok(NULL, delimiter);
				if (ptr != NULL)
				{
					if (0 == strcasecmp_P(ptr, PSTR("true")))
					{
						serialLog = true;
						serialport_printf("OK\r");
						serialport_printf("Timestp;tarTemp;actTemp;Devi;p;i;d;actuatingV\r");
					}
					else if (0 == strcasecmp_P(ptr, PSTR("false")))
					{
						serialLog = false;
						serialport_printf("OK\r");
					}
					else
					{
						serialport_printf("ERROR, true or false required\r");
					}
				}
				else
				{
					serialport_printf("ERROR, true or false required\r");
				}
			}
			else if (0 == strcasecmp_P(ptr, PSTR("help")))				// integrated help function
			{
				serialport_printf("OK\r");
				serialport_printf("actual; get actual iron temperature\r");
				serialport_printf("log <true/false>; enable controller logging\r");
				serialport_printf("reset; reset settings and reboot to bootloader\r");
				serialport_printf("standby <true/false>; standby temperature on/off\r");
				serialport_printf("target <value>; set iron temperature\r");
				serialport_printf("version; print version string\r");
			} 
			else
			{
				serialport_printf("ERROR, unknown command, try help\r");
			}
		}						
	}	// end while
}		// end main


uint8_t calc_pid(uint16_t targetTemperature, float actualTemperature, bool logEnable) // execute the pid controller with 1,0 Hz
{
	// intermediate calculation values:
	static uint16_t logTimestamp = 0;
	static float actuatingVariable = 0;
	static float temperatureDeviation_old = 0;
	static float iPart = 0;
	float temperatureDeviation;
	float pPart;
	float dPart;
	
	// calculate temperature deviation
	temperatureDeviation = (float)targetTemperature - actualTemperature;
	
	// calculate integral part:
	if (actuatingVariable < 100)                                          // here: actuating variable of last cycle
	{                                                                     // in controllable region
		iPart += temperatureDeviation * i_value;                          // adding deviations
		if (iPart > 100) iPart = 100;                                     // and limit i part to 0...100
		if (iPart < 0) iPart = 0;
	}
	else
	{                                                                     // full power, out of control
		iPart = 0;                                                        // no integral part
	}
	
	// calculate differential part
	dPart = (temperatureDeviation - temperatureDeviation_old) * d_value;  // difference of last and actual deviation
	if (dPart > 100) dPart = 100;                                         // limit to -50...100
	if (dPart < -50) dPart = -50;
	temperatureDeviation_old = temperatureDeviation;                      // save actual deviation as old for the next cycle
	
	// calculate proportional part
	pPart = temperatureDeviation * p_value;                               // That's easy ;-)
	
	// overall sum
	actuatingVariable = pPart + iPart + dPart;                                 
	if (actuatingVariable > 100) actuatingVariable = 100;                 // and limit to 0...100 percent
	if (actuatingVariable < 0) actuatingVariable = 0;
	
	// log if required
	if (logEnable)
	{
		serialport_printf("%05u;%3u;%5.1f;%+6.1f;%4.0f;%3.0f;%3.0f;%3.0f\r"
							,logTimestamp
							,targetTemperature
							,actualTemperature
							,temperatureDeviation
							,pPart
							,iPart
							,dPart
							,actuatingVariable );
		logTimestamp++;
	}
	else
	{
		logTimestamp = 0;
	}
	
	// give result as byte
	return (uint8_t)actuatingVariable;
}

void settings_menu(void)
{
	uint8_t menu_idx = 1;
	int8_t contrast_percent = (int8_t)( ( ( (int16_t)contrast * 100 ) + 128 ) / 255 );
	
	pwm_set_power(0);	//switch power off during menu set
	
	ssd1309_clear();
	ssd1309_print_P(PSTR("SETTINGS:"), 0, 8);
	ssd1309_print_P(PSTR("V"), 0, 70);
	ssd1309_print_P(PSTR(VERSION), 0, 80);
	ssd1309_print_P(PSTR("Contrast :"), 1, 8);
	ssd1309_print_P(PSTR("STBY-Temp:"), 2, 8);
	ssd1309_print_P(PSTR("STBY-Time:"), 3, 8);
	ssd1309_print_P(PSTR("Tool-Type:"), 4, 8);
	ssd1309_print_P(PSTR("P-Val-SET:"), 5, 8);
	ssd1309_print_P(PSTR("I-Val-SET:"), 6, 8);
	ssd1309_print_P(PSTR("D-Val-SET:"), 7, 8);
	ssd1309_picture(arrowUp, 1, 1, 121, 6);
	ssd1309_picture(arrowDown, 7, 7, 121, 6);
	ssd1309_picture(arrowExit, 4, 4, 121, 6);	
	
	ssd1309_picture(marker, menu_idx, menu_idx, 0, 6);
	
	while (1)
	{
		if (timer_10ms_tick())
		{
			gpio_poll();							// poll switches
		}	
		
		if (gpio_checkButton(BUTTON_1, SHORT))
		{
			ssd1309_print_P(PSTR(" "), menu_idx, 0);
			
			if (menu_idx > 1)
			{
				menu_idx--;
			}	
			else
			{
				menu_idx = 7;
			}			
		
			ssd1309_picture(marker, menu_idx, menu_idx, 0, 6);
						
		}
		
		if (gpio_checkButton(BUTTON_ROT_ENC, SHORT))
		{
			ssd1309_print_P(PSTR(" "), menu_idx, 0);
			
			if (menu_idx < 7)
			{
				menu_idx++;
			}
			else
			{
				menu_idx = 1;
			}
						
			ssd1309_picture(marker, menu_idx, menu_idx, 0, 6);
			
		}
		
		if (gpio_checkButton(BUTTON_2, SHORT))
		{
			break;
		}
		
		if (gpio_checkButton(BUTTON_3, SHORT))
		{
			ssd1309_print_P(PSTR(" "), menu_idx, 0);
			
			if (menu_idx < 7)
			{
				menu_idx++;
			}
			else
			{
				menu_idx = 1;
			}
			
			ssd1309_picture(marker, menu_idx, menu_idx, 0, 6);
			
		}
		
		ssd1309_printf_P(1, 70, PSTR("%3i%%"), contrast_percent);
		ssd1309_printf_P(2, 70, PSTR("%3iC"), standby_temp);
		ssd1309_printf_P(3, 70, PSTR("%2umin"), standby_time);
		// draw tool name
		switch (tool)
		{
		case WELLER_WP80:
			ssd1309_print_P(PSTR("WP80 "), 4, 70);
			break;
		case WELLER_WSP80:
			ssd1309_print_P(PSTR("WSP80"), 4, 70);
			break;
		default:
			break;
		}
		ssd1309_printf_P(5, 70, PSTR("%6.3f"), p_value);
		ssd1309_printf_P(6, 70, PSTR("%6.3f"), i_value);
		ssd1309_printf_P(7, 70, PSTR("%6.3f"), d_value);
		
		int16_t dummy_time;
		
 		switch (menu_idx)
 		{			 
		case 1:
			contrast_percent += gpio_getRotEnc() * 10;
			
			if (contrast_percent < 0)
			{
				contrast_percent = 0;
			}
			if (contrast_percent > 100)
			{
				contrast_percent = 100;
			}
			contrast = (uint8_t)( ( ( (int16_t)contrast_percent * 255 ) + 50 ) / 100 );
			ssd1309_contrast(contrast);
			break;
			
		case 2:
 			standby_temp += gpio_getRotEnc();
 			if (standby_temp < 0)
 			{
	 			standby_temp = 0;
 			}
 			if (standby_temp > 300)
 			{
	 			standby_temp = 300;
 			}
 			break;
			 
 		case 3:
			dummy_time = standby_time;
			dummy_time += gpio_getRotEnc();
			if (dummy_time < 5)
			{
				dummy_time = 5;
			}
			if (dummy_time > 90)
			{
				dummy_time = 90;
			}
			standby_time = dummy_time;
			break;
			
		case 4:
			tool += gpio_getRotEnc();
			if (tool >= TOOL_LAST)
			{
				tool = 0;
			}
			break;
		
		case 5:
			p_value += ((float)gpio_getRotEnc() / 40.0);
			if (p_value < 0.0)
			{
				p_value = 0.0;
			}
			if (p_value > 25.0)
			{
				p_value = 25.0;
			}
			break;
			
		case 6:
			i_value += ((float)gpio_getRotEnc() / 200.0);
			if (i_value < 0.0)
			{
				i_value = 0.0;
			}
			if (i_value > 2.0)
			{
				i_value = 2.0;
			}
			break;
			
		case 7:
			d_value += ((float)gpio_getRotEnc() / 40.0);
			if (d_value < 0.0)
			{
				d_value = 0.0;
			}
			if (d_value > 25.0)
			{
				d_value = 25.0;
			}
			break;
		 }
	}	//end while
	
	ssd1309_clear();
	
	eeprom_update_block(&tool, &eeprom_tool, sizeof(tool));
	eeprom_update_byte(&eeprom_contrast, contrast);
	eeprom_update_block(&standby_temp, &eeprom_standby_temp, sizeof(standby_temp));
	eeprom_update_byte(&eeprom_standby_time, standby_time);
	eeprom_update_float(&eeprom_p_value, p_value);
	eeprom_update_float(&eeprom_i_value, i_value);
	eeprom_update_float(&eeprom_d_value, d_value);
	
}		//end function