/**
 * @file	tempsens.c
 * @author	Robert Steigemann
 * @date	15 January 2017
 * @brief	Measurement library for temperature sensor
 *
 * @license creative commons license CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/
 */

#include <avr/io.h>
#include <stddef.h>
#include <math.h>
#include <avr/pgmspace.h>

#include "tempsens.h"

// AD-converter parameter
#define TEMPSENS_GAIN		   8L
#define TEMPSENS_TOP		2047L
#define TEMPSENS_UREF		   1L
#define TEMPSENS_R56		(TEMPSENS_R5+TEMPSENS_R6)
#define TEMPSENS_R567		(TEMPSENS_R5+TEMPSENS_R6+TEMPSENS_R7)

// temperature profile for Weller WP80 tool
#define TEMPSENS_WP80_A	     0.00535
#define TEMPSENS_WP80_B	    -0.00000205
#define TEMPSENS_WP80_R0	18.5

// temperature profile for Weller WSP80 tool
#define TEMPSENS_WSP80_A     0.0058
#define TEMPSENS_WSP80_B    -0.0000024
#define TEMPSENS_WSP80_R0   20.0


// low pass filter buffer and index
static uint16_t lpBuffer[TEMPSENS_LP_SIZE] = {UINT16_MAX};	// holds Rsens in mOhm
static uint8_t lpBufferIdx = 0;


static uint8_t read_calibration_byte(int idx)
{
	uint8_t result;	
	NVM_CMD = NVM_CMD_READ_CALIB_ROW_gc;
	result = pgm_read_byte(idx);
	NVM_CMD = NVM_CMD_NO_OPERATION_gc;
	return(result);
}


void tempsens_init(void)
{
	// read and set factory calibration values
	ADCA.CALL =	read_calibration_byte(offsetof(NVM_PROD_SIGNATURES_t, ADCACAL0));
	ADCA.CALH =	read_calibration_byte(offsetof(NVM_PROD_SIGNATURES_t, ADCACAL1));												
	
	ADCA_CTRLB     = ADC_RESOLUTION_12BIT_gc | ADC_CONMODE_bm;		// 12 Bit and signed mode
	ADCA_REFCTRL   = ADC_REFSEL_INT1V_gc | ADC_BANDGAP_bm;			// 1.00 V internal band gap reference
	ADCA_PRESCALER = ADC_PRESCALER_DIV512_gc;						// prescaler is 512 resulting in 62.5 kHz
	ADCA_CTRLA     = ADC_ENABLE_bm;									// enable ADC
	
	// Channel 0 for VCC measurement
	ADCA_CH0_CTRL = ADC_CH_INPUTMODE_SINGLEENDED_gc;				
	ADCA_CH0_MUXCTRL = (4 << 3);									// connected to ADC4
	
	// Channel 1 for R Value measurement
	ADCA_CH1_CTRL = ADC_CH_INPUTMODE_DIFFWGAIN_gc | ADC_CH_GAIN_8X_gc;
	ADCA_CH1_MUXCTRL = (6 << 3) | (4 - 4);							// connected to ADC6 and ADC4	
}


tempsensState_t tempsens_measure(void)
{
	// start both channels
	ADCA_CH0_CTRL |= ADC_CH_START_bm;
	ADCA_CH1_CTRL |= ADC_CH_START_bm;
	
	// wait until both channels are finished
	while(!ADCA_CH0_INTFLAGS);
	while(!ADCA_CH1_INTFLAGS);
	
	int32_t vccVal = (int16_t)ADCA_CH0_RES;
	int32_t rVal = (int16_t)ADCA_CH1_RES;
	
	//                vccVal     R5 + R6 + R7
	// AVCC = Uref x -------- x --------------
	//                 TOP            R7
	int32_t AVCC = (vccVal*TEMPSENS_R567*10)/TEMPSENS_TOP;	// [E-1]
	        AVCC = (AVCC*1000)/TEMPSENS_R7;					// [E-4]
	
	if (AVCC<31000 || AVCC>35000)	// AVCC is out of range, vccVal 472-532
	{
		return AVCC_OUT_OF_RANGE;		
	}
	
	if (rVal == TEMPSENS_TOP)		// no sensor connected
	{
		return TOOL_DISCONNECTED;
	}
		
	//          Uref       rVal
	// Usens = ------ x ( ------ + vccVal )
	//          TOP        GAIN
	int32_t Usens = ((rVal+(vccVal*TEMPSENS_GAIN))*100000) / (TEMPSENS_TOP*TEMPSENS_GAIN);	// [E-5 V]	
	
	//      AVCC - Usens
	// I = --------------
	//          R8
	int32_t I = ((AVCC*10-Usens))/TEMPSENS_R8;	// [E-5 A]
	
	//          Usens
	// Rsens = -------
	//            I
	int32_t Rsens = (Usens*100)/I;	// [E-2 Ohm]
	
	// limit Rsens
	if (Rsens < 0)
	{
		Rsens = 0;
	} 
	else if (Rsens > UINT16_MAX)
	{
		Rsens = UINT16_MAX;
	}
	
	// increment buffer index
	lpBufferIdx++;
	if (lpBufferIdx >= TEMPSENS_LP_SIZE)
	{
		lpBufferIdx = 0;
	}
		
	// put the measured resistance value in low pass filter buffer
	lpBuffer[lpBufferIdx] = (uint16_t)Rsens;
	
	return OK;	
}


float tempsens_get(solderType_t solderTool)
{
	uint32_t sum = 0;
	uint8_t idx;
	float t, r;
	
	// calculate mid value
	for (idx=0; idx<TEMPSENS_LP_SIZE; idx++)
	{
		sum += lpBuffer[idx];
	}
	
	r = (float)sum / (100.0*TEMPSENS_LP_SIZE);
	
	// Rsens = R0 x ( 1 + at + bt )
	//         a             a          r - R0
	// t = - ---- - sqrt[ (----)^2 + ------------ ]
	//        2b            2b          b x R0	
	switch (solderTool)
	{
	default:
	case WELLER_WP80:
		t = -(TEMPSENS_WP80_A/(2*TEMPSENS_WP80_B)) - sqrt((TEMPSENS_WP80_A/(2*TEMPSENS_WP80_B))*(TEMPSENS_WP80_A/(2*TEMPSENS_WP80_B)) + (r-TEMPSENS_WP80_R0)/(TEMPSENS_WP80_B*TEMPSENS_WP80_R0));
		break;
		
	case WELLER_WSP80:
		t = -(TEMPSENS_WSP80_A/(2*TEMPSENS_WSP80_B)) - sqrt((TEMPSENS_WSP80_A/(2*TEMPSENS_WSP80_B))*(TEMPSENS_WSP80_A/(2*TEMPSENS_WSP80_B)) + (r-TEMPSENS_WSP80_R0)/(TEMPSENS_WSP80_B*TEMPSENS_WSP80_R0));
		break;		
	}	
	
	if (t < 0.0)			// limit temperature to 0..500.0 C
	{
		t = 0.0;
	} 
	else if (t > 500.0)
	{
		t = 500.0;
	} 
	
	return t;
}